/* $Id$

This file is part of gldbg
Copyright © 2009-2010  Mikko Rasa, Mikkosoft Productions
Distributed under the GPL
*/

#include <cstdlib>
#include <fcntl.h>
#include <signal.h>
#include <sys/poll.h>
#include <sys/socket.h>
#include <readline/readline.h>
#include <msp/core/except.h>
#include <msp/fs/dir.h>
#include <msp/io/print.h>
#include <msp/strings/lexicalcast.h>
#include "functions.h"
#include "gldbg.h"
#include "gldecoder.h"
#include "packet.h"
#include "tool.h"

using namespace std;
using namespace Msp;

Application::RegApp<GlDbg> GlDbg::reg;

GlDbg::GlDbg(int argc, char **argv):
	cmd_interp(*this),
	process(vector<string>(argv+1, argv+argc)),
	buf_offset(0),
	flushing(false),
	got_sigchld(false)
{
	FS::Path libdir = FS::get_sys_lib_dir(argv[0], "gldbg");
	process.setenv("LD_PRELOAD", (libdir/"glwrap.so").str().c_str());

	const list<Tool::Factory *> &factories = Tool::get_factories();
	for(list<Tool::Factory *>::const_iterator i=factories.begin(); i!=factories.end(); ++i)
		tools.push_back((*i)->create(*this));
}

GlDbg::~GlDbg()
{
	for(list<Tool *>::iterator i=tools.begin(); i!=tools.end(); ++i)
		delete *i;
}

int GlDbg::main()
{
	catch_signal(SIGINT);
	catch_signal(SIGCHLD);
	set_loop_mode(TICK_BUSY);

	IO::print("GLdbg 0.0\n");
	IO::print("Copyright © 2009-2010 Mikkosoft Productions\n");
	IO::print("Type \"help\" for a list of commands\n");

	Application::main();

	return 0;
}

void GlDbg::launch()
{
	if(process.get_state()!=Process::INACTIVE)
		throw InvalidState("Program is already running");

	int fds[2];
	socketpair(AF_UNIX, SOCK_STREAM, 0, fds);
	sock_fd = fds[0];

	int flags = fcntl(sock_fd, F_GETFD);
	fcntl(sock_fd, F_SETFD, flags|FD_CLOEXEC);

	process.setenv("GLWRAP_FD", lexical_cast(fds[1]));
	process.setenv("GLWRAP_CTRL_FD", lexical_cast(fds[1]));
	process.launch();
	close(fds[1]);
}

void GlDbg::set_breakpoint(unsigned short func, char flag)
{
	GlPacket *pkt = packet_begin(FUNC_GLDBREAK);
	packet_write_short(pkt, func);
	packet_write_char(pkt, flag);
	packet_write_char(pkt, 0);
	packet_send(pkt, sock_fd);
}

void GlDbg::clear_breakpoint(unsigned short func, char flag)
{
	GlPacket *pkt = packet_begin(FUNC_GLDBREAK);
	packet_write_short(pkt, func);
	packet_write_char(pkt, 0);
	packet_write_char(pkt, flag);
	packet_send(pkt, sock_fd);
}

void GlDbg::quit(bool force)
{
	if(!force && process.get_state()!=Process::INACTIVE)
		throw InvalidState("Program is still running");
	exit(0);
}

void GlDbg::tick()
{
	if(got_sigchld)
	{
		got_sigchld = false;
		int ret = process.check();
		if(ret==0x100)
			IO::print("Target process exited normally\n");
		else if((ret>>8)==1)
			IO::print("Target process exited with status %d\n", ret&0xFF);
		else if((ret>>8)==2)
			IO::print("Target process terminated with signal %d\n", ret&0xFF);
		else if((ret>>8)==3)
		{
			IO::print("Target process stopped by signal %d\n", ret&0xFF);
			flushing = true;
		}
	}

	Process::State pstate = process.get_state();
	if((pstate!=Process::INACTIVE && pstate!=Process::STOPPED) || flushing)
		read_stream();
	else
	{
		char *line = readline("gldbg> ");
		if(line)
		{
			try
			{
				cmd_interp.execute(line);
			}
			catch(const Exception &e)
			{
				IO::print("%s\n", e.what());
			}
			free(line);
		}
		else if(pstate==Process::INACTIVE)
			exit(0);
	}
}

void GlDbg::read_stream()
{
	pollfd pfd = { sock_fd, POLLIN, 0 };
	int ret = poll(&pfd, 1, (flushing ? 0 : -1));
	if(ret>0)
	{
		char rbuf[1024];
		ret = read(sock_fd, rbuf, 1024);
		if(ret>0)
		{
			buffer.append(rbuf, ret);
			while(buffer.size()>buf_offset)
			{
				const char *data = buffer.data()+buf_offset;
				unsigned len = buffer.size()-buf_offset;
				GlPacket *pkt = packet_receive_str(data, &len);
				if(!pkt)
					break;

				unsigned short func;
				packet_read_short(pkt, (short *)&func);
				if(func==FUNC_GLDBREAK)
				{
					packet_read_short(pkt, (short *)&func);
					IO::print("Breakpoint: %s\n", get_function_name(func));
				}

				for(list<Tool *>::iterator i=tools.begin(); i!=tools.end(); ++i)
					(*i)->decode(data, len);
				buf_offset += len;
			}
			if(buf_offset>8192)
			{
				buffer.erase(0, buf_offset);
				buf_offset = 0;
			}
		}
	}
	else if(flushing)
		flushing = false;
}

void GlDbg::sighandler(int sig)
{
	if(sig==SIGCHLD)
		got_sigchld = true;
}
